@@ -7,12 +7,17 @@ from .cython_generator import (
 )
 from .ext_module import ExtModule
 from .extern import Extern
-from .low_level import Kernel, LocalMem, Cython
+from .low_level import Kernel, LocalMem, Cython, cast
 from .parallel import (
     Elementwise, Reduction, Scan, elementwise
 )
+from .profile import (
+    get_profile_info, named_profile, profile, profile_ctx, print_profile,
+    profile_kernel, ProfileContext, profile2csv
+)
 from .translator import (
     CConverter, CStructHelper, OpenCLConverter, detect_type, ocl_detect_type,
     py2c
 )
 from .types import KnownType, annotate, declare
+from .utils import ArgumentParser
