@@ -11,7 +11,7 @@ from .config import get_config
 from .cython_generator import CythonGenerator
 from .transpiler import Transpiler, BUILTINS
 from .types import (dtype_to_ctype, get_declare_info,
-                    dtype_to_knowntype, annotate)
+                    dtype_to_knowntype, annotate, BITS, KnownType)
 from .extern import Extern
 from .utils import getsourcelines
 from .profile import profile
@@ -34,30 +34,31 @@ def memoize_kernel(key=lambda *args: args):
     return memoize_deco
 
 
-def get_ctype_from_arg(arg):
+def get_ctype_from_arg(arg, backend=None):
     if isinstance(arg, array.Array):
         return arg.gptr_type
     elif isinstance(arg, np.ndarray) or isinstance(arg, np.floating):
-        return dtype_to_ctype(arg.dtype)
+        return dtype_to_ctype(arg.dtype, backend=backend)
     else:
         if isinstance(arg, float):
             return 'double'
         else:
             if arg > 2147483648:
-                return 'long'
+                return 'long long' if BITS.startswith('32') else 'long'
             else:
                 return 'int'
 
 
 def kernel_cache_key_args(obj, *args):
-    key = [get_ctype_from_arg(arg) for arg in args]
+    key = [get_ctype_from_arg(arg, backend=obj.backend) for arg in args]
     key.append(obj.func)
     key.append(obj.name)
     return tuple(key + list(parallel.get_common_cache_key(obj)))
 
 
 def kernel_cache_key_kwargs(obj, **kwargs):
-    key = [get_ctype_from_arg(arg) for arg in kwargs.values()]
+    key = [get_ctype_from_arg(arg, backend=obj.backend)
+           for arg in kwargs.values()]
     key.append(obj.input_func)
     key.append(obj.output_func)
     key.append(obj.scan_expr)
@@ -65,15 +66,28 @@ def kernel_cache_key_kwargs(obj, **kwargs):
 
 
 def getargspec(f):
-    getargspec_f = getattr(inspect, 'getfullargspec',
-                           getattr(inspect, 'getargspec'))
+    getargspec_f = inspect.getfullargspec
     return getargspec_f(f)[0]
 
 
+def get_signed_type(a):
+    return a[1:] if a.startswith('u') else a
+
+
 def get_binop_return_type(a, b):
+    int_types = ['short', 'int', 'long']
+    float_types = ['float', 'double']
+
     if a is None or b is None:
         return None
-    preference_order = ['short', 'int', 'long', 'float', 'double']
+
+    if a.endswith('p') and get_signed_type(b) in int_types:
+        return a
+    if b.endswith('p') and get_signed_type(a) in int_types:
+        return b
+
+    preference_order = int_types + float_types
+
     unsigned_a = unsigned_b = False
     if a.startswith('u'):
         unsigned_a = True
@@ -81,6 +95,7 @@ def get_binop_return_type(a, b):
     if b.startswith('u'):
         unsigned_b = True
         b = b[1:]
+
     idx_a = preference_order.index(a)
     idx_b = preference_order.index(b)
     return_type = preference_order[idx_a] if idx_a > idx_b else \
@@ -93,8 +108,10 @@ def get_binop_return_type(a, b):
 class AnnotationHelper(ast.NodeVisitor):
     def __init__(self, func, arg_types):
         self.func = func
-        self.arg_types = arg_types
-        self.var_types = arg_types.copy()
+        self.name = self.func.__name__
+        self.arg_types = {name: self.get_declare_type(type_str)
+                          for name, type_str in arg_types.items()}
+        self.var_types = self.arg_types.copy()
         self.undecl_var_types = {}
         self.external_funcs = {}
         self.external_missing_decl = {}
@@ -102,10 +119,9 @@ class AnnotationHelper(ast.NodeVisitor):
             Function called is not marked by the annotate decorator. Argument
             type defaulting to 'double'. If the type is not 'double', store
             the value in a variable of appropriate type and use the variable
-            '''
-                            )
+        ''')
 
-    def get_type(self, type_str):
+    def get_declare_type(self, type_str):
         kind, address_space, ctype, shape = get_declare_info(type_str)
         if 'unsigned' in ctype:
             ctype = ctype.replace('unsigned ', 'u')
@@ -121,14 +137,46 @@ class AnnotationHelper(ast.NodeVisitor):
         missing_decl.update(self.external_missing_decl)
         return missing_decl
 
+    def record_var_type(self, name, dtype):
+        self.var_types[name] = self.get_declare_type(dtype)
+
+    def record_undecl_var_type(self, name, dtype):
+        if name not in self.var_types and name not in self.undecl_var_types:
+            self.undecl_var_types[name] = self.get_declare_type(dtype)
+
+    def get_var_type(self, name):
+        return self.var_types.get(
+            name, self.undecl_var_types.get(name, 'double'))
+
+    def get_return_type(self):
+        if getattr(self.func, 'is_jit', False):
+            return self.arg_types.get('return_', 'double')
+        else:
+            annotations = getattr(self.func, '__annotations__', {})
+            return annotations.get('return', KnownType('double')).type
+
     def annotate(self):
-        src = dedent('\n'.join(getsourcelines(self.func)[0]))
-        self._src = src.splitlines()
-        code = ast.parse(src)
-        self.visit(code)
-        self.func = annotate(self.func, **self.arg_types)
+        if getattr(self.func, 'is_jit', False):
+            src = dedent('\n'.join(getsourcelines(self.func)[0]))
+            self._src = src.splitlines()
+            code = ast.parse(src)
+            self.visit(code)
+            self.func = annotate(self.func, **self.arg_types)
         return self.get_missing_declarations(self.undecl_var_types)
 
+    def recursive_annotate(self, f, node):
+        arg_types = {}
+        f_arg_names = getargspec(f)
+        for f_arg, arg in zip(f_arg_names, node.args):
+            arg_type = self.visit(arg)
+            if not arg_type:
+                arg_type = 'double'
+            arg_types[f_arg] = arg_type
+        f_helper = AnnotationHelper(f, arg_types)
+        self.external_missing_decl.update(f_helper.annotate())
+        self.external_funcs[node.func.id] = f_helper
+        return f_helper
+
     def error(self, message, node):
         msg = '\nError in code in line %d:\n' % node.lineno
         if self._src:  # pragma: no branch
@@ -149,10 +197,29 @@ class AnnotationHelper(ast.NodeVisitor):
         msg += message
         warnings.warn(msg)
 
+    def visit_declare(self, node):
+        if not isinstance(node.args[0], ast.Str):
+            self.error("Argument to declare should be a string.", node)
+        type_str = node.args[0].s
+        return self.get_declare_type(type_str)
+
+    def visit_cast(self, node):
+        if not isinstance(node.args[1], ast.Str):
+            self.error("Cast type should be a string.", node)
+        return node.args[1].s
+
+    def visit_address(self, node):
+        base_type = self.visit(node.args[0])
+        if base_type.endswith('p'):
+            self.error("Cannot find address of a pointer", node)
+        if isinstance(node.args[0], ast.Subscript):
+            array_type = self.visit(node.args[0].value)
+            if array_type.startswith('g'):
+                base_type = 'g' + base_type
+        return base_type + 'p'
+
     def visit_For(self, node):
-        if node.target.id not in self.var_types and \
-                node.target.id not in self.undecl_var_types:
-            self.undecl_var_types[node.target.id] = 'int'
+        self.record_undecl_var_type(node.target.id, 'int')
         for stmt in node.body:
             self.visit(stmt)
 
@@ -163,32 +230,32 @@ class AnnotationHelper(ast.NodeVisitor):
         # FIXME: External functions have to be at the module level
         # for this to work. Pass list of external functions to
         # make this work
+        if node.func.id == 'annotate':
+            return
         mod = importlib.import_module(self.func.__module__)
         f = getattr(mod, node.func.id, None)
-        if node.func.id not in BUILTINS and not hasattr(f, 'is_jit'):
-            return None
+        if node.func.id == 'declare':
+            return self.visit_declare(node)
+        if node.func.id == 'cast':
+            return self.visit_cast(node)
+        if node.func.id == 'atomic_inc':
+            return self.visit(node.args[0])
+        if node.func.id == 'atomic_dec':
+            return self.visit(node.args[0])
+        if node.func.id == 'address':
+            return self.visit_address(node)
         if node.func.id in self.external_funcs:
-            return self.external_funcs[node.func.id].arg_types.get(
-                'return_', None)
-        if isinstance(node.func, ast.Name) and \
-                node.func.id not in BUILTINS:
+            return self.external_funcs[node.func.id].get_return_type()
+        if isinstance(node.func, ast.Name) and node.func.id not in BUILTINS:
             if f is None or isinstance(f, Extern):
-                return None
+                self.warn("%s could not be found or is an external function"
+                          "and cannot be handled by JIT" % node.func.id)
+                return 'double'
             else:
-                arg_types = []
-                for arg in node.args:
-                    arg_type = self.visit(arg)
-                    if not arg_type:
-                        self.warn(dedent(self.warning_msg), arg)
-                        arg_type = 'double'
-                    arg_types.append(arg_type)
-                # make a new helper and call visit
-                f_arg_names = getargspec(f)
-                f_arg_types = dict(zip(f_arg_names, arg_types))
-                f_helper = AnnotationHelper(f, f_arg_types)
-                self.external_missing_decl.update(f_helper.annotate())
-                self.external_funcs[node.func.id] = f_helper
-                return f_helper.arg_types.get('return_', None)
+                f_helper = self.recursive_annotate(f, node)
+                return f_helper.get_return_type()
+        self.warn(dedent(self.warning_msg), node.func)
+        return 'double'
 
     def visit_Subscript(self, node):
         base_type = self.visit(node.value)
@@ -197,49 +264,22 @@ class AnnotationHelper(ast.NodeVisitor):
         return base_type[:-1]
 
     def visit_Name(self, node):
-        node_type = self.var_types.get(
-            node.id, self.undecl_var_types.get(node.id, 'double')
-        )
-        return node_type
+        return self.get_var_type(node.id)
 
     def visit_Assign(self, node):
-        # Only for declare calls
         if len(node.targets) != 1:
             self.error("Assignments can have only one target.", node)
         left, right = node.targets[0], node.value
-        if isinstance(right, ast.Call) and isinstance(right.func, ast.Name):
-            if right.func.id == 'declare':
-                if not isinstance(right.args[0], ast.Str):
-                    self.error("Argument to declare should be a string.", node)
-                type = right.args[0].s
-                if isinstance(left, ast.Name):
-                    self.var_types[left.id] = self.get_type(type)
-                elif isinstance(left, ast.Tuple):
-                    names = [x.id for x in left.elts]
-                    for name in names:
-                        self.var_types[name] = self.get_type(type)
-            elif right.func.id == 'cast':
-                if not isinstance(right.args[1], ast.Str):
-                    self.error("Cast type should be a string.", node)
-                type = right.args[1].s
-                if isinstance(left, ast.Name):
-                    self.undecl_var_types[left.id] = self.get_type(type)
-            elif right.func.id == 'atomic_inc':
-                if left.id not in self.var_types and \
-                        left.id not in self.undecl_var_types:
-                    self.undecl_var_types[left.id] = self.visit(right.args[0])
-            elif isinstance(left, ast.Name):
-                if left.id not in self.var_types and \
-                        left.id not in self.undecl_var_types:
-                    self.undecl_var_types[left.id] = self.visit(right)
-                else:
-                    self.visit(right)
+        right_type = self.visit(right)
+        if isinstance(right, ast.Call) and right.func.id == 'declare':
+            if isinstance(left, ast.Name):
+                self.record_var_type(left.id, right_type)
+            elif isinstance(left, ast.Tuple):
+                names = [x.id for x in left.elts]
+                for name in names:
+                    self.record_var_type(name, right_type)
         elif isinstance(left, ast.Name):
-            if left.id not in self.var_types and \
-                    left.id not in self.undecl_var_types:
-                self.undecl_var_types[left.id] = self.visit(right)
-            else:
-                self.visit(right)
+            self.record_undecl_var_type(left.id, right_type)
 
     def visit_Compare(self, node):
         return 'int'
@@ -252,36 +292,17 @@ class AnnotationHelper(ast.NodeVisitor):
                                          self.visit(node.right))
 
     def visit_Num(self, node):
-        if isinstance(node.n, float):
-            return_type = 'double'
-        else:
-            if node.n > 2147483648:
-                return_type = 'long'
-            else:
-                return_type = 'int'
-        return return_type
+        return get_ctype_from_arg(node.n)
 
     def visit_UnaryOp(self, node):
         return self.visit(node.operand)
 
     def visit_Return(self, node):
-        if isinstance(node.value, ast.Name) or \
-                isinstance(node.value, ast.Subscript) or \
-                isinstance(node.value, ast.Num) or \
-                isinstance(node.value, ast.BinOp) or \
-                isinstance(node.value, ast.Call) or \
-                isinstance(node.value, ast.IfExp) or \
-                isinstance(node.value, ast.UnaryOp):
+        if node and node.value:
             result_type = self.visit(node.value)
             if result_type:
                 self.arg_types['return_'] = result_type
-            else:
-                self.arg_types['return_'] = 'double'
-        else:
-            if node.value:
-                self.warn("Unknown type for return value. "
-                          "Return value defaulting to 'double'", node)
-                self.arg_types['return_'] = 'double'
+                return result_type
 
 
 class ElementwiseJIT(parallel.ElementwiseBase):
@@ -306,7 +327,7 @@ class ElementwiseJIT(parallel.ElementwiseBase):
             arg_names.remove('i')
             type_info['i'] = 'int'
         for arg, name in zip(args, arg_names):
-            arg_type = get_ctype_from_arg(arg)
+            arg_type = get_ctype_from_arg(arg, backend=self.backend)
             if not arg_type:
                 arg_type = 'double'
             type_info[name] = arg_type
@@ -351,7 +372,7 @@ class ElementwiseJIT(parallel.ElementwiseBase):
 
 class ReductionJIT(parallel.ReductionBase):
     def __init__(self, reduce_expr, map_func=None, dtype_out=np.float64,
-                 neutral='0', backend='cython'):
+                 neutral='0', backend=None):
         backend = array.get_backend(backend)
         self.tp = Transpiler(backend=backend)
         self.backend = backend
@@ -362,7 +383,7 @@ class ReductionJIT(parallel.ReductionBase):
             self.name = 'reduce'
         self.reduce_expr = reduce_expr
         self.dtype_out = dtype_out
-        self.type = dtype_to_ctype(dtype_out)
+        self.type = dtype_to_ctype(dtype_out, backend)
         if backend == 'cython':
             # On Windows, INFINITY is not defined so we use INFTY which we
             # internally define.
@@ -384,7 +405,7 @@ class ReductionJIT(parallel.ReductionBase):
             arg_names.remove('i')
             type_info['i'] = 'int'
         for arg, name in zip(args, arg_names):
-            arg_type = get_ctype_from_arg(arg)
+            arg_type = get_ctype_from_arg(arg, backend=self.backend)
             if not arg_type:
                 arg_type = 'double'
             type_info[name] = arg_type
@@ -432,7 +453,7 @@ class ReductionJIT(parallel.ReductionBase):
 class ScanJIT(parallel.ScanBase):
     def __init__(self, input=None, output=None, scan_expr="a+b",
                  is_segment=None, dtype=np.float64, neutral='0',
-                 complex_map=False, backend='opencl'):
+                 complex_map=False, backend=None):
         backend = array.get_backend(backend)
         self.tp = Transpiler(backend=backend, incl_cluda=False)
         self.backend = backend
@@ -446,7 +467,7 @@ class ScanJIT(parallel.ScanBase):
             self.name = 'scan'
         self.scan_expr = scan_expr
         self.dtype = dtype
-        self.type = dtype_to_ctype(dtype)
+        self.type = dtype_to_ctype(dtype, backend)
         if backend == 'cython':
             # On Windows, INFINITY is not defined so we use INFTY which we
             # internally define.
@@ -463,7 +484,9 @@ class ScanJIT(parallel.ScanBase):
         builtin_symbols = ['item', 'prev_item', 'last_item']
         self.builtin_types = {'i': 'int', 'N': 'int'}
         for sym in builtin_symbols:
-            self.builtin_types[sym] = dtype_to_knowntype(self.dtype)
+            self.builtin_types[sym] = dtype_to_knowntype(
+                self.dtype, backend=backend
+            )
 
     def get_type_info_from_kwargs(self, func, **kwargs):
         type_info = {}
@@ -475,7 +498,7 @@ class ScanJIT(parallel.ScanBase):
             if name in self.builtin_types:
                 arg_type = self.builtin_types[name]
             else:
-                arg_type = get_ctype_from_arg(arg)
+                arg_type = get_ctype_from_arg(arg, backend=self.backend)
             if not arg_type:
                 arg_type = 'double'
             type_info[name] = arg_type
@@ -487,7 +510,9 @@ class ScanJIT(parallel.ScanBase):
         if self.input_func is not None:
             arg_types = self.get_type_info_from_kwargs(
                 self.input_func, **kwargs)
-            arg_types['return_'] = dtype_to_knowntype(self.dtype)
+            arg_types['return_'] = dtype_to_knowntype(
+                self.dtype, backend=self.backend
+            )
             helper = AnnotationHelper(self.input_func, arg_types)
             declarations.update(helper.annotate())
             self.input_func = helper.func
@@ -523,7 +548,7 @@ class ScanJIT(parallel.ScanBase):
         c_args_dict = {k: self._massage_arg(x) for k, x in kwargs.items()}
         if self._get_backend_key() in self.output_func.arg_keys:
             output_arg_keys = self.output_func.arg_keys[
-                    self._get_backend_key()]
+                self._get_backend_key()]
         else:
             raise ValueError("No kernel arguments found for backend = %s, "
                              "use_openmp = %s, use_double = %s" %
