@@ -153,7 +153,7 @@ class Kernel(object):
         return re.sub(r'\bdouble\b', 'float', s)
 
     def _get_func_info(self):
-        getfullargspec = getattr(inspect, 'getfullargspec', inspect.getargspec)
+        getfullargspec = inspect.getfullargspec
         argspec = getfullargspec(self.func)
         annotations = getattr(
             argspec, 'annotations', self.func.__annotations__
@@ -321,6 +321,14 @@ class _atomic_inc(Extern):
         pass
 
 
+class _atomic_dec(Extern):
+    def code(self, backend):
+        return ''
+
+    def __call__(self, *args, **kw):
+        pass
+
+
 class _cast(Extern):
     def code(self, backend):
         return ''
@@ -334,6 +342,7 @@ parallel = _parallel()
 nogil = _nogil()
 address = _address()
 atomic_inc = _atomic_inc()
+atomic_dec = _atomic_dec()
 cast = _cast()
 
 
