@@ -29,11 +29,11 @@ cdef c_${name}(${c_arg_sig}):
 %if openmp:
     with nogil, parallel():
         for i in ${get_parallel_range("SIZE")}:
-%else:
-    if 1:
-        for i in range(SIZE):
-%endif
             ${name}(${c_args})
+%else:
+    for i in range(SIZE):
+        ${name}(${c_args})
+%endif
 
 cpdef py_${name}(${py_arg_sig}):
     c_${name}(${py_args})
@@ -556,6 +556,13 @@ class ElementwiseBase(object):
 
 class Elementwise(object):
     def __init__(self, func, backend=None):
+        self._func = func
+        self._backend = backend
+        self.elementwise = None
+
+    def _setup(self):
+        func = self._func
+        backend = array.get_backend(self._backend)
         if getattr(func, '__annotations__',
                    None) and not hasattr(func, 'is_jit'):
             self.elementwise = ElementwiseBase(func, backend=backend)
@@ -570,8 +577,14 @@ class Elementwise(object):
         return sorted(dir(self.elementwise) + ['elementwise'])
 
     def __call__(self, *args, **kwargs):
+        if self.elementwise is None:
+            self._setup()
         self.elementwise(*args, **kwargs)
 
+    def set_backend(self, backend=None):
+        self._backend = backend
+        self.elementwise = None
+
 
 def elementwise(func=None, backend=None):
     def _wrapper(function):
@@ -585,7 +598,7 @@ def elementwise(func=None, backend=None):
 
 class ReductionBase(object):
     def __init__(self, reduce_expr, map_func=None, dtype_out=np.float64,
-                 neutral='0', backend='cython'):
+                 neutral='0', backend=None):
         backend = array.get_backend(backend)
         self.tp = Transpiler(backend=backend)
         self.backend = backend
@@ -596,7 +609,7 @@ class ReductionBase(object):
             self.name = 'reduce'
         self.reduce_expr = reduce_expr
         self.dtype_out = dtype_out
-        self.type = dtype_to_ctype(dtype_out)
+        self.type = dtype_to_ctype(dtype_out, backend=backend)
         if backend == 'cython':
             # On Windows, INFINITY is not defined so we use INFTY which we
             # internally define.
@@ -742,7 +755,7 @@ class ReductionBase(object):
     def _correct_return_type(self, c_data):
         code = self.tp.blocks[-1].code.splitlines()
         if self._config.use_openmp:
-            gil = " nogil"
+            gil = " noexcept nogil"
         else:
             gil = ""
         code[0] = "cdef inline {type} {name}({args}){gil}:".format(
@@ -806,7 +819,23 @@ class ReductionBase(object):
 
 class Reduction(object):
     def __init__(self, reduce_expr, map_func=None, dtype_out=np.float64,
-                 neutral='0', backend='cython'):
+                 neutral='0', backend=None):
+        self._reduce_expr = reduce_expr
+        self._map_func = map_func
+        self._dtype_out = dtype_out
+        self._neutral = neutral
+        self._backend = backend
+        self.reduction = None
+
+    def _setup(self):
+        map_func = self._map_func
+        reduce_expr = self._reduce_expr
+        dtype_out, neutral = self._dtype_out, self._neutral
+        backend = array.get_backend(self._backend)
+        cfg = get_config()
+        if (backend != 'cython' and dtype_out == np.float64 and
+           not cfg.use_double):
+            dtype_out = np.float32
         if map_func is None or getattr(map_func, '__annotations__', None) and \
                 not hasattr(map_func, 'is_jit'):
             self.reduction = ReductionBase(reduce_expr, map_func=map_func,
@@ -827,8 +856,14 @@ class Reduction(object):
         return getattr(self.reduction, name)
 
     def __call__(self, *args, **kwargs):
+        if self.reduction is None:
+            self._setup()
         return self.reduction(*args, **kwargs)
 
+    def set_backend(self, backend=None):
+        self._backend = backend
+        self.reduction = None
+
 
 class ScanBase(object):
     def __init__(self, input=None, output=None, scan_expr="a+b",
@@ -847,7 +882,7 @@ class ScanBase(object):
             self.name = 'scan'
         self.scan_expr = scan_expr
         self.dtype = dtype
-        self.type = dtype_to_ctype(dtype)
+        self.type = dtype_to_ctype(dtype, backend=backend)
         if backend == 'cython':
             # On Windows, INFINITY is not defined so we use INFTY which we
             # internally define.
@@ -869,7 +904,7 @@ class ScanBase(object):
     def _correct_return_type(self, c_data, modifier):
         code = self.tp.blocks[-1].code.splitlines()
         if self._config.use_openmp:
-            gil = " nogil"
+            gil = " noexcept nogil"
         else:
             gil = ""
         code[0] = "cdef inline {type} {name}_{modifier}({args}){gil}:".format(
@@ -925,7 +960,7 @@ class ScanBase(object):
             expr = '{name}_{modifier}({cargs})'.format(name=name, cargs=cargs,
                                                        modifier=func_type)
         else:
-            if func_type is 'input':
+            if func_type == 'input':
                 py_data, c_data, expr = self._default_cython_input_function()
             else:
                 py_data, c_data, expr = [], [], None
@@ -1032,7 +1067,7 @@ class ScanBase(object):
             arguments = self._filter_ignored(c_data[0], select)
             c_args = self._filter_ignored(c_data[1], select)
         else:
-            if func_type is 'input':
+            if func_type == 'input':
                 if self.backend == 'opencl':
                     arguments = ['__global %(type)s *input' %
                                  {'type': self.type}]
@@ -1147,7 +1182,7 @@ class ScanBase(object):
             return arg
 
     def _correct_opencl_address_space(self, c_data, func, func_type):
-        return_type = 'void' if func_type is 'output' else self.type
+        return_type = 'void' if func_type == 'output' else self.type
         code = self.tp.blocks[-1].code.splitlines()
         header_idx = 1
         for line in code:
@@ -1205,6 +1240,21 @@ class Scan(object):
     def __init__(self, input=None, output=None, scan_expr="a+b",
                  is_segment=None, dtype=np.float64, neutral='0',
                  complex_map=False, backend=None):
+        self._input = input
+        self._output = output
+        self._scan_expr = scan_expr
+        self._is_segment = is_segment
+        self._dtype = dtype
+        self._neutral = neutral
+        self._complex_map = complex_map
+        self._backend = backend
+        self.scan = None
+
+    def _setup(self):
+        input, output, scan_expr = self._input, self._output, self._scan_expr
+        is_segment, dtype = self._is_segment, self._dtype
+        neutral, complex_map = self._neutral, self._complex_map
+        backend = array.get_backend(self._backend)
         # FIXME: Revisit these conditions
         input_base = input is None or \
             getattr(input, '__annotations__', None) and \
@@ -1239,4 +1289,10 @@ class Scan(object):
         return getattr(self.scan, name)
 
     def __call__(self, **kwargs):
+        if self.scan is None:
+            self._setup()
         self.scan(**kwargs)
+
+    def set_backend(self, backend=None):
+        self._backend = backend
+        self.scan = None
