@@ -3,11 +3,11 @@ import pytest
 import numpy as np
 import sys
 
-from ..config import get_config
+from ..config import get_config, use_config
 from ..types import annotate, declare
 from ..translator import (
     CConverter, CodeGenerationError, CStructHelper, KnownType,
-    OpenCLConverter, CUDAConverter, py2c
+    OpenCLConverter, CUDAConverter, literal_to_float, py2c
 )
 
 
@@ -17,22 +17,50 @@ def annotated_f(i, y):
     return y[i]
 
 
+def test_literal_to_float():
+    # Given/When/Then
+    assert literal_to_float(1.0) == '1.0f'
+    assert literal_to_float(1.01325e5) == '101325.0f'
+    assert literal_to_float(1.01325e-5) == '1.01325e-05f'
+    assert literal_to_float(1.01325E5) == '101325.0f'
+    assert literal_to_float(1.01325e-05) == '1.01325e-05f'
+    assert literal_to_float(1.0e-5) == '1e-05f'
+    assert literal_to_float(1.0, use_double=True) == '1.0'
+    assert literal_to_float(10, use_double=True) == '10'
+    assert literal_to_float(1.01325e5, use_double=True) == '101325.0'
+    assert literal_to_float(1.01325e-5, use_double=True) == '1.01325e-05'
+
+
 def test_simple_assignment_expression():
     # Given
     src = dedent('''
     b = (2*a + 1)*(-a/1.5)%2
     ''')
 
-    # When
-    code = py2c(src)
+    with use_config(use_double=True):
+        # When
+        code = py2c(src)
 
-    # Then
-    expect = dedent('''
-    double a;
-    double b;
-    b = ((((2 * a) + 1) * (-a / 1.5)) % 2);
-    ''')
-    assert code == expect.strip()
+        # Then
+        expect = dedent('''
+        double a;
+        double b;
+        b = ((((2 * a) + 1) * (-a / 1.5)) % 2);
+        ''')
+        assert code == expect.strip()
+
+    # Given
+    with use_config(use_double=False):
+        # When
+        code = py2c(src)
+
+        # Then
+        expect = dedent('''
+        double a;
+        double b;
+        b = ((((2 * a) + 1) * (-a / 1.5f)) % 2);
+        ''')
+        assert code == expect.strip()
 
 
 def test_multiple_assignment_expressions():
@@ -49,11 +77,24 @@ def test_multiple_assignment_expressions():
     expect = dedent('''
     double a;
     double b;
-    a = 21.5;
-    b = ((((2 * a) + 1) * (a / 1.5)) % 2);
+    a = 21.5f;
+    b = ((((2 * a) + 1) * (a / 1.5f)) % 2);
     ''')
     assert code == expect.strip()
 
+    with use_config(use_double=True):
+        # When
+        code = py2c(src)
+
+        # Then
+        expect = dedent('''
+        double a;
+        double b;
+        a = 21.5;
+        b = ((((2 * a) + 1) * (a / 1.5)) % 2);
+        ''')
+        assert code == expect.strip()
+
 
 def test_if_block():
     # Given
@@ -74,7 +115,7 @@ def test_if_block():
     expect = dedent('''
     double a;
     double b;
-    a = 21.5;
+    a = 21.5f;
     if ((a > 20)) {
         b = (a - 1);
     }
@@ -154,17 +195,31 @@ def test_ternary_operator():
     x = 1.0 if y >= 2.0 else 0.0
     ''')
 
-    # When
-    code = py2c(src)
+    with use_config(use_double=True):
+        # When
+        code = py2c(src)
 
-    # Then
-    expect = dedent('''
-    double x;
-    double y;
-    y = 2.0;
-    x = (y >= 2.0) ? 1.0 : 0.0;
-    ''')
-    assert code.strip() == expect.strip()
+        # Then
+        expect = dedent('''
+        double x;
+        double y;
+        y = 2.0;
+        x = (y >= 2.0) ? 1.0 : 0.0;
+        ''')
+        assert code.strip() == expect.strip()
+
+    with use_config(use_double=False):
+        # When
+        code = py2c(src)
+
+        # Then
+        expect = dedent('''
+        double x;
+        double y;
+        y = 2.0f;
+        x = (y >= 2.0f) ? 1.0f : 0.0f;
+        ''')
+        assert code.strip() == expect.strip()
 
 
 def test_multiple_boolops():
@@ -223,7 +278,7 @@ def test_power():
     # Then
     expect = dedent('''
     double x;
-    (1.5 * pow(x, 2));
+    (1.5f * pow(x, 2));
     ''')
     assert code.strip() == expect.strip()
 
@@ -251,7 +306,7 @@ def test_calling_function():
 
     # Then
     expect = dedent('''
-    sin((23.2 + 1));
+    sin((23.2f + 1));
     ''')
     assert code == expect.strip()
 
@@ -267,7 +322,7 @@ def test_calling_printf_with_string():
 
     # Then
     expect = dedent('''
-    printf("%s %d %f\n", "hello", 1, 2.0);
+    printf("%s %d %f\n", "hello", 1, 2.0f);
     ''')
     assert code == expect.strip()
 
@@ -741,15 +796,15 @@ def test_for_with_symbols():
         ;
     }
 
-    __cpy_stop_0 = (n + 1);
-    long __cpy_step_0 = step();
-    if (__cpy_step_0 < 0) {
-        for (long i=0; i>__cpy_stop_0; i+=__cpy_step_0) {
+    long __cpy_stop_1 = (n + 1);
+    long __cpy_step_1 = step();
+    if (__cpy_step_1 < 0) {
+        for (long i=0; i>__cpy_stop_1; i+=__cpy_step_1) {
             ;
         }
     }
     else {
-        for (long i=0; i<__cpy_stop_0; i+=__cpy_step_0) {
+        for (long i=0; i<__cpy_stop_1; i+=__cpy_step_1) {
             ;
         }
     }
@@ -793,17 +848,17 @@ def test_nested_for_with_symbols():
         }
     }
 
-    __cpy_stop_0 = (n + 1);
-    for (long i=0; i<__cpy_stop_0; i+=1) {
-        long __cpy_stop_1 = (n + 2);
-        long __cpy_step_1 = step();
-        if (__cpy_step_1 < 0) {
-            for (long j=0; j>__cpy_stop_1; j+=__cpy_step_1) {
+    long __cpy_stop_2 = (n + 1);
+    for (long i=0; i<__cpy_stop_2; i+=1) {
+        long __cpy_stop_3 = (n + 2);
+        long __cpy_step_3 = step();
+        if (__cpy_step_3 < 0) {
+            for (long j=0; j>__cpy_stop_3; j+=__cpy_step_3) {
                 ;
             }
         }
         else {
-            for (long j=0; j<__cpy_stop_1; j+=__cpy_step_1) {
+            for (long j=0; j<__cpy_stop_3; j+=__cpy_step_3) {
                 ;
             }
         }
@@ -812,6 +867,60 @@ def test_nested_for_with_symbols():
     assert code.strip() == expect.strip()
 
 
+def test_with_two_functions():
+    # Given
+    src = dedent('''
+    def f():
+        n = declare('int')
+        n = 20
+        for i in range(n):
+            pass
+        for i in range(n):
+            pass
+    def g():
+        n = declare('int')
+        n = 20
+        for i in range(n):
+            pass
+        for i in range(n):
+            pass
+    ''')
+    code = py2c(src)
+
+    expect = dedent('''
+    void f()
+    {
+        int n;
+        n = 20;
+        long __cpy_stop_0 = n;
+        for (long i=0; i<__cpy_stop_0; i+=1) {
+            ;
+        }
+        long __cpy_stop_1 = n;
+        for (long i=0; i<__cpy_stop_1; i+=1) {
+            ;
+        }
+    }
+
+    void g()
+    {
+        int n;
+        n = 20;
+        long __cpy_stop_0 = n;
+        for (long i=0; i<__cpy_stop_0; i+=1) {
+            ;
+        }
+        long __cpy_stop_1 = n;
+        for (long i=0; i<__cpy_stop_1; i+=1) {
+            ;
+        }
+    }
+
+
+    ''')
+    assert code.strip() == expect.strip()
+
+
 def test_for_with_break_continue():
     # Given
     src = dedent('''
@@ -1022,25 +1131,47 @@ def test_class():
     ''')
 
     # When
-    code = py2c(src)
+    with use_config(use_double=True):
+        code = py2c(src)
 
-    # Then
-    expect = dedent('''
-    double Foo_g(Foo* self, double x)
-    {
-        return (x * 2.0);
-    }
+        # Then
+        expect = dedent('''
+        double Foo_g(Foo* self, double x)
+        {
+            return (x * 2.0);
+        }
 
-    void Foo_f(Foo* self, double x)
-    {
-        double y;
-        double z;
-        y = (x + 1);
-        do(self->a, x);
-        z = Foo_g(self, y);
-    }
-    ''')
-    assert code.strip() == expect.strip()
+        void Foo_f(Foo* self, double x)
+        {
+            double y;
+            double z;
+            y = (x + 1);
+            do(self->a, x);
+            z = Foo_g(self, y);
+        }
+        ''')
+        assert code.strip() == expect.strip()
+
+    with use_config(use_double=False):
+        code = py2c(src)
+
+        # Then
+        expect = dedent('''
+        double Foo_g(Foo* self, double x)
+        {
+            return (x * 2.0f);
+        }
+
+        void Foo_f(Foo* self, double x)
+        {
+            double y;
+            double z;
+            y = (x + 1);
+            do(self->a, x);
+            z = Foo_g(self, y);
+        }
+        ''')
+        assert code.strip() == expect.strip()
 
 
 def test_unsupported_method():
@@ -1303,17 +1434,18 @@ def test_handles_parsing_functions():
             return x + 1.0
 
     # When
-    t = CConverter()
-    code = t.parse_function(A)
+    with use_config(use_double=False):
+        t = CConverter()
+        code = t.parse_function(A)
 
-    # Then
-    expect = dedent('''
-    double A_f(A* self, double x)
-    {
-        return (x + 1.0);
-    }
-    ''')
-    assert code.strip() == expect.strip()
+        # Then
+        expect = dedent('''
+        double A_f(A* self, double x)
+        {
+            return (x + 1.0f);
+        }
+        ''')
+        assert code.strip() == expect.strip()
 
 
 def test_address_works():
@@ -1335,7 +1467,7 @@ def test_address_works():
     assert code.strip() == expect.strip()
 
 
-def test_atomic_works():
+def test_atomic_inc_works():
     # Given
     def f(x=1.0):
         return atomic_inc(x)
@@ -1368,6 +1500,41 @@ def test_atomic_works():
 
     assert code.strip() == expect.strip()
 
+
+def test_atomic_dec_works():
+    # Given
+    def f(x=1.0):
+        return atomic_dec(x)
+
+    # When
+    t = OpenCLConverter()
+    code = t.parse_function(f)
+
+    # Then
+    expect = dedent('''
+    WITHIN_KERNEL double f(double x)
+    {
+        return atomic_dec(&x);
+    }
+    ''')
+
+    assert code.strip() == expect.strip()
+
+    # When
+    t = CUDAConverter()
+    code = t.parse_function(f)
+
+    # Then
+    expect = dedent('''
+    WITHIN_KERNEL double f(double x)
+    {
+        return atomicAdd(&x, -1);
+    }
+    ''')
+
+    assert code.strip() == expect.strip()
+
+
 def test_cast_works():
     # Given
     def f(x=1.0):
