@@ -30,6 +30,15 @@ from .utils import getsource
 PY_VER = sys.version_info.major
 
 
+def literal_to_float(value, use_double=False):
+    """Convert numerical value to a string in 32 bit floating point notation.
+    """
+    if isinstance(value, float) and not use_double:
+        return str(value) + 'f'
+    else:
+        return str(value)
+
+
 def detect_type(name, value):
     if isinstance(value, KnownType):
         return value.type
@@ -122,6 +131,7 @@ class CStructHelper(object):
 
 class CConverter(ast.NodeVisitor):
     def __init__(self, detect_type=detect_type, known_types=None):
+        self._use_double = get_config().use_double
         self._declares = {}
         self._known = set((
             'M_E', 'M_LOG2E', 'M_LOG10E', 'M_LN2', 'M_LN10',
@@ -553,7 +563,6 @@ class CConverter(ast.NodeVisitor):
                     block=block
                 )
 
-            self._for_count -= 1
             if count == 0:
                 self._known -= self._added_loop_vars
                 self._added_loop_vars = set()
@@ -572,6 +581,7 @@ class CConverter(ast.NodeVisitor):
                                  node.name in self._ignore_methods):
             return ''
 
+        self._for_count = 0
         orig_declares = self._declares
         self._declares = {} if not self._declarations else self._declarations
         orig_known = set(self._known)
@@ -679,7 +689,7 @@ class CConverter(ast.NodeVisitor):
         return '!='
 
     def visit_Num(self, node):
-        return str(node.n)
+        return literal_to_float(node.n, self._use_double)
 
     def visit_Or(self, node):
         return '||'
@@ -700,8 +710,12 @@ class CConverter(ast.NodeVisitor):
         return r'"%s"' % node.s
 
     def visit_Subscript(self, node):
+        if sys.version_info >= (3, 9):
+            sliceval = node.slice
+        else:
+            sliceval = node.slice.value
         return '%s[%s]' % (
-            self.visit(node.value), self.visit(node.slice.value)
+            self.visit(node.value), self.visit(sliceval)
         )
 
     def visit_TryExcept(self, node):
@@ -777,8 +791,11 @@ class OpenCLConverter(CConverter):
     def render_atomic(self, func, arg):
         if func == 'atomic_inc':
             return 'atomic_inc(&%s)' % self.visit(arg)
+        elif func == 'atomic_dec':
+            return 'atomic_dec(&%s)' % self.visit(arg)
         else:
-            raise NotImplementedError("Only atomic_inc supported right now")
+            raise NotImplementedError(
+                "Only atomic_inc, atomic_dec supported right now")
 
 
 class CUDAConverter(OpenCLConverter):
@@ -827,5 +844,8 @@ class CUDAConverter(OpenCLConverter):
     def render_atomic(self, func, arg):
         if func == 'atomic_inc':
             return 'atomicAdd(&%s, 1)' % self.visit(arg)
+        elif func == 'atomic_dec':
+            return 'atomicAdd(&%s, -1)' % self.visit(arg)
         else:
-            raise NotImplementedError("Only atomic_inc supported right now")
+            raise NotImplementedError(
+                "Only atomic_inc, atomic_dec supported right now")
